/******************************************************************************
 * JibsGame.java - Everything about a single game of backgammon on jIBS.
 * $Id: JibsGame.java,v 1.13 2011/06/08 20:53:52 dick Exp $
 * 
 * jIBS - A Java clone of FIBS
 * Copyright 2002 - Axel Leucht
 * Copyright 2011 - Dick Balaska - BuckoSoft, Corp.
 * 
 * $Log: JibsGame.java,v $
 * Revision 1.13  2011/06/08 20:53:52  dick
 * constructGame(resumeData) becomes a constructor.
 * Still playing with greedy bearoffs.
 *
 * Revision 1.12  2011/06/07 19:49:17  dick
 * onHome/Bar 1/2 becomes onHomeBar X/O
 * outBoard() takes a Player, not a JibsWriter.  -- We need to know what style the Player wants.
 * Finish crawford attributes.
 * Don't be greedy if !isRace().
 * roll[OX]() takes 2 players, not a player and two streams.
 * Cleanup forced move/greedy handling.
 *
 * Revision 1.11  2011/06/06 03:18:26  dick
 * JibsMessages belongs in n.s.jibs.server, not n.s.jibs.gui
 *
 * Revision 1.10  2011/06/06 02:45:54  dick
 * getPlayer() comes from JibsServer now.
 *
 * Revision 1.9  2011/06/05 18:26:34  dick
 * log is a method of JibsServer.  Always write screen logs to the log file.
 * So, no need to pass write-to-logs=true from everywhere.
 *
 * Revision 1.8  2011/06/05 05:56:58  dick
 * Get rid of PlayerA and PlayerB in the constructor.  PlayerX and O will do just fine.
 * Get rid of Player1 and 2, and MayDouble1 and 2.  Call them PlayerX and PlayerO and MayDoubleX and O.
 *
 * Revision 1.7  2011/06/03 19:25:52  dick
 * Rework winGame handling.  Now there is just one function regardless of X or O winning.
 * Fix backgammon times cube calculation.
 *
 * Revision 1.6  2011/06/03 04:14:49  dick
 * Better message generation for winning game messages.
 *
 * Revision 1.5  2011/06/03 00:25:09  dick
 * Player.show2Watcher() becomes show2Watchers().
 *
 * Revision 1.4  2011/06/02 19:30:11  dick
 * informWatcher() becomes informWatchers().
 *
 * Revision 1.3  2011/05/31 13:02:26  dick
 * Generics cleanups.
 *
 * Revision 1.2  2011/05/30 05:29:21  dick
 * retab.
 *
 */

/* 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The Original Code is jIBS, <https://sourceforge.net/projects/fibs/>.
 * The Initial Developer of the Original Code is Axel Leucht.
 * 
 */
package net.sourceforge.jibs.backgammon;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.StringTokenizer;

import net.sourceforge.jibs.server.JibsMessages;
import net.sourceforge.jibs.server.JibsServer;
import net.sourceforge.jibs.server.Player;
import net.sourceforge.jibs.util.GameEnder;
import net.sourceforge.jibs.util.JibsWriter;
import net.sourceforge.jibs.util.RatingCalculator;
import net.sourceforge.jibs.util.SavedGameParam;

import org.apache.log4j.Logger;

import com.ibatis.sqlmap.client.SqlMapClient;

/**
 * This models one game between two players. A match, hence can consist of
 * several games.
 *
 * @author Axel Leucht
 * @author Dick Balaska
 * @version $Revision: 1.13 $ <br> $Date: 2011/06/08 20:53:52 $
 * @see <a href="http://cvs.buckosoft.com/Projects/jIBS/jIBS/src/main/java/net/sourceforge/jibs/backgammon/JibsGame.java">cvs JibsGame.java</a>
 */
public class JibsGame {
	private static final Logger logger = Logger.getLogger(JibsGame.class);
	private BackgammonBoard board;
	private Collection<PossibleMove> col;
	private JibsServer jibsServer;
	private Player playerX;
	private Player playerO;

	/** Default constructor */
	public JibsGame() {
	}

	/** Construct a JibsGame with these parameters
	 * @param jibsServer The jibsServer
	 * @param playerX Who is playing PlayerX
	 * @param playerO Who is playing PlayerO
	 * @param matchlength How many points in the match
	 * @param playerXPoints How many points does X have
	 * @param playerOPoints How many points does O have
	 * @param matchVersion {@link JibsMatch#nPointMatch} or {@link JibsMatch.unlimitedMatch} 
	 */
	public JibsGame(JibsServer jibsServer, Player playerX, Player playerO,
			int matchlength,
			int playerXPoints, int playerOPoints, JibsMatch matchVersion) {
		this.jibsServer = jibsServer;
		this.playerX = playerX;
		this.playerO = playerO;
		this.board = new BackgammonBoard(jibsServer, playerX,
				playerO, matchlength,
				matchVersion);
	}

	public JibsGame (JibsServer jibsServer,
			SavedGameParam resumeData) {
		Player playerX = resumeData.getPlayerX();
		Player playerO = resumeData.getPlayerO();

		BackgammonBoard board = new BackgammonBoard(jibsServer,
				playerX,
				playerO,
				resumeData.getMatchlength()
				.intValue(),
				new JibsMatch(resumeData.getMatchVersion()));
		board.setPlayerX(playerX);
		board.setPlayerO(playerO);
		this.setBackgammonBoard(board);
		this.getBackgammonBoard()
		.setMatchVersion(new JibsMatch(resumeData.getMatchVersion()));

		Die d1 = new Die(jibsServer);
		Die d2 = new Die(jibsServer);

		this.getBackgammonBoard().setpDie1(d1);
		this.getBackgammonBoard().setpDie2(d2);

		StringTokenizer tokenizer = new StringTokenizer(resumeData.getBoard(),
				":");
		int position = 0;
		boolean bError = false;
		int help;

		while (tokenizer.hasMoreTokens()) {
			String x = tokenizer.nextToken();

			switch (position) {
			case 0:

				if (!x.equals("board")) {
					bError = true;
				}
				break;

			case 1:
				board.setPlayerX(playerX);
				board.setPlayerXName(playerX.getName());

				if (!x.equals("You")) {
					bError = true;
				}
				break;

			case 2:

				Player p = jibsServer.getPlayer(x);
				board.setPlayerOName(p.getName());
				board.setPlayerO(p);
				break;

			case 3:
				help = Integer.valueOf(x);
				board.setMatchlength(help);
				break;

			case 4:
				help = Integer.valueOf(x);
				board.setPlayerXScore(help);
				break;

			case 5:
				help = Integer.valueOf(x);
				board.setPlayerOScore(help);
				break;

			case 6:
			case 7:
			case 8:
			case 9:
			case 10:
			case 11:
			case 12:
			case 13:
			case 14:
			case 15:
			case 16:
			case 17:
			case 18:
			case 19:
			case 20:
			case 21:
			case 22:
			case 23:
			case 24:
			case 25:
			case 26:
			case 27:
			case 28:
			case 29:
			case 30:
			case 31:
				help = Integer.valueOf(x);
				board.getBoard()[position - 6] = help;

				break;

			case 32: // turn

			int turn = Integer.valueOf(x);
			board.setTurn(turn);

			break;

			case 33: // player dice
			help = Integer.valueOf(x);
			board.getpDie1().setValue(help);

			break;

			case 34: // player dice
			help = Integer.valueOf(x);
			board.getpDie2().setValue(help);

			break;

			case 35: // opp dice
			help = Integer.valueOf(x);
			board.getoDie1().setValue(help);

			break;

			case 36: // opp dice
				help = Integer.valueOf(x);
				board.getoDie2().setValue(help);

				break;

			case 37: // cube
				help = Integer.valueOf(x);
				board.setCubeNumber(help);

				break;

			case 38: // maydouble1
				help = Integer.valueOf(x);
				board.setMayDoubleX(help);

				break;

			case 39: // maydouble2
				help = Integer.valueOf(x);
				board.setMayDoubleO(help);

				break;

			case 40: // wasdoubled
				help = Integer.valueOf(x);
				board.setWasDoubled(help);

				break;

			case 41: // color
				help = Integer.valueOf(x);
				board.setColor(help);

				break;

			case 42: // direction
				help = Integer.valueOf(x);
				board.setDirection(help);

				break;

			case 43: // home
				help = Integer.valueOf(x);
				board.setHome(help);

				break;

			case 44: // bar
				help = Integer.valueOf(x);
				board.setBar(help);

				break;

			case 45: // player onHome
				help = Integer.valueOf(x);
				board.setOnHomeX(help);

				break;

			case 46: // opp onHome
				help = Integer.valueOf(x);
				board.setOnHomeO(help);

				break;

			case 47: // player onBar
				help = new Integer(x).intValue();
				board.setOnBarX(help);

				break;

			case 48: // opp onBar
				help = Integer.valueOf(x);
				board.setOnBarO(help);

				break;

			case 49: // canMove
				help = Integer.valueOf(x);
				board.setCanMove(help);

				break;

			case 50: // forced move
				help = Integer.valueOf(x);
				board.setForcedMove(help);

				break;

			case 51: // did crawford
				help = Integer.valueOf(x);
				board.setDidCrawford(help);

				break;

			case 52: // redoubles
				help = new Integer(x).intValue();
				board.setRedoubles(help);

				break;
			}

			position++;
		}

		if (tokenizer.hasMoreTokens()) {
			bError = true;
		}
		if (bError) {
			logger.error("Failed to restore game from database");
			logger.error(resumeData.getBoard());
		}
	}

	public static void deleteGames(JibsServer jibsServer, Player player1,
			Player player2) {
		try {
			SavedGameParam savedGame = new SavedGameParam(player1.getName(),
					player2.getName());
			SqlMapClient sqlClient = jibsServer.getSqlMapInstance();
			sqlClient.delete("SavedGames.deleteSavedGame", savedGame);
		} catch (SQLException e) {
			jibsServer.logException(e);
		}
	}

	public static SavedGameParam loadGame(JibsServer jibsServer, String name1,
			String name2) {
		SqlMapClient sqlMapClient = jibsServer.getSqlMapInstance();
		SavedGameParam savedGame = new SavedGameParam(name1, name2);
		SavedGameParam retSavedGame = null;

		try {
			retSavedGame = (SavedGameParam) sqlMapClient.queryForObject("SavedGames.loadSavedGame",
					savedGame);
		} catch (SQLException e) {
			jibsServer.logException(e);
		}

		return retSavedGame;
	}

	/** Calculate the possible moves and output a board
	 * @param game
	 * @param player
	 * @param isFirstMove
	 * @param turn
	 * @param nrChances
	 * @param playerX
	 * @param playerO
	 * @param board
	 * @param dice1
	 * @param dice2
	 * @return
	 */
	private int calcPossibleMovesAndOutBoard(JibsGame game, Player player,
			boolean isFirstMove, int turn,
			MoveChances nrChances, Player playerX,
			Player playerO, BackgammonBoard board,
			int dice1, int dice2) {
		Object[] obj = null;
		String msg = null;
		int nrMoves = 0;
		Player opponent = board.getOpponent(player);

		switch (turn) {
		case -1: // X

			BackgammonBoard opBoard = board.switch2O();
			opBoard.setCanMove(0);
			opBoard.outBoard(playerO, "You", turn, 0, 0, dice1, dice2); // O
			// out2.println("board:You:aleucht:2:0:0:0:-2:0:0:0:0:5:0:3:0:0:0:-5:5:0:0:0:-3:0:-5:0:0:0:0:2:0:-1:0:0:5:3:1:1:1:0:1:-1:0:25:0:0:0:0:0:0:0:0");
			// out2.println("board:You:aleucht:2:0:1:0:-2:0:0:0:0:5:0:3:0:0:0:-5:5:0:0:0:-3:0:-5:0:0:0:0:2:0:-1:0:0:5:3:1:1:0:0:1:-1:0:25:0:0:0:0:0:0:0:0");

			board.getPlayerO()
			.show2WatcherBoard(opBoard, board.getPlayerO().getName(), 0,
					0, dice1, dice2);
			nrMoves = nrChances.calcPossibleMovesX(turn, dice1, dice2);
			board.setCanMove(nrMoves);
			board.outBoard(playerX, "You", turn, dice1, dice2, 0, 0); // X
			board.getPlayerX()
			.show2WatcherBoard(board, board.getPlayerX().getName(), dice1,
					dice2, 0, 0);

			if (!isFirstMove) {
				if (nrMoves > 0) {
					// m_please_move=Please move %0 pieces.
					obj = new Object[] { Integer.valueOf(nrMoves) };
					msg = jibsServer.getJibsMessages()
					.convert("m_please_move", obj);
					player.getOutputStream().println(msg);
				} else {
					// m_you_cant_move=You can't move.
					msg = jibsServer.getJibsMessages().convert("m_you_cant_move");
					player.getOutputStream().println(msg);
					// m_other_cant_move=%0 can't move.
					obj = new Object[] { player.getName() };
					msg = jibsServer.getJibsMessages()
					.convert("m_other_cant_move", obj);
					opponent.getOutputStream().println(msg);
				}
			}

			break;

		case 1: // O
			nrMoves = nrChances.calcPossibleMovesO(turn, dice1, dice2);
			board.setCanMove(0);
			board.outBoard(playerX, "You", turn, 0, 0, dice1, dice2); // X
			board.getPlayerX()
			.show2WatcherBoard(board, board.getPlayerX().getName(), 0, 0,
					dice1, dice2);
			board.setCanMove(nrMoves);
			opBoard = board.switch2O();
			opBoard.outBoard(playerO, "You", turn, dice1, dice2, 0, 0); // O
			board.getPlayerO()
			.show2WatcherBoard(opBoard, board.getPlayerO().getName(),
					dice1, dice2, 0, 0);

			if (!isFirstMove) {
				if (nrMoves > 0) {
					// m_please_move=Please move %0 pieces.
					obj = new Object[] { Integer.valueOf(nrMoves) };
					msg = jibsServer.getJibsMessages()
					.convert("m_please_move", obj);
					playerO.println(msg);
				} else {
					// m_you_cant_move=You can't move.
					msg = jibsServer.getJibsMessages().convert("m_you_cant_move");
					playerO.println(msg);
					// m_other_cant_move=%0 can't move.
					obj = new Object[] { board.getPlayerO().getName() };
					msg = jibsServer.getJibsMessages()
					.convert("m_other_cant_move", obj);
					playerX.println(msg);
				}
			}

			break;
		}

		/*
		 * System.err.println("----------------------------"); Iterator iter =
		 * col.iterator(); while(iter.hasNext()) { PossibleMove pMove =
		 * (PossibleMove) iter.next(); System.err.println(pMove.toString()); }
		 */
		return nrMoves;
	}

	public String checkForcedMove(Player player, BackgammonBoard bgBoard,
			int nrMoves) {
		JibsMessages jibsMessages = jibsServer.getJibsMessages();

		if (player.checkToggle("automove")) {
			if (col.size() == 1) {
				Iterator<PossibleMove> iter = col.iterator();
				PossibleMove posMove = iter.next();
				String movString = posMove.conv2MoveString();

				// m_only_move=The only possible move is %0.
				Object[] obj = new Object[] { movString };
				String msg = jibsMessages.convert("m_only_move", obj);
				player.println(msg);
				logger.info(msg);
				String nextCmd = "m " + movString;

				return nextCmd;
			}
		}

		return null;
	}

	public String checkGreedy(Player player, int nrMoves) {
		if (player.checkToggle("greedy") && board.isRace()) {
			if (col.size() > 0) {
/*				for (PossibleMove posMove : col) {
					if (posMove.getnrMoves() == nrMoves) {
						logger.info("Got greedy move = \"" + posMove.conv2MoveString() + "\"");
						return("m " + posMove.conv2MoveString());
					}
				}
*/

				BackgammonBoard helpboard = new BackgammonBoard(board);

				for (PossibleMove posMove : col) {
					BackgammonBoard board1 = null;
					BackgammonBoard setBoard = new BackgammonBoard(board);

					for (int j = 0; j < posMove.getnrMoves(); j++) {
						Move mv = posMove.getMove(j);

						if (mv != null) {
							board1 = setBoard.placeMoveX(mv);
							setBoard = board1;
						}
					}

					int diff = setBoard.getOnHomeX() - helpboard.getOnHomeX();

					if (diff >= nrMoves) {
						// yes, we've found a possible greedy move
						return "m " + posMove.conv2MoveString();
					}
				}
			}

		}

		return null;
	}

	public BackgammonBoard getBackgammonBoard() {
		return board;
	}

	public void resumeGame(SavedGameParam resumeData, int startTurn, Die dice1,
			Die dice2, int player1Points, int player2Points) {
		SavedGameParam savedGame = new SavedGameParam(resumeData.getPlayer_A(),
				resumeData.getPlayer_B());
		SqlMapClient sqlClient = jibsServer.getSqlMapInstance();

		try {
			sqlClient.delete("SavedGames.deleteSavedGame", savedGame);
		} catch (SQLException e) {
			jibsServer.logException(e);
		}

		Player playerX = resumeData.getPlayerX();
		Player playerO = resumeData.getPlayerO();

		if (board.getMatchVersion().getVersion() == JibsMatch.nPointMatch) {
			// m_resume_match=%0 and %1 are resuming their %2-point match.
			Object[] obj = {
					playerX.getName(), playerO.getName(),
					Integer.valueOf(resumeData.getMatchlength())
			};
			JibsMessages jibsMessages = jibsServer.getJibsMessages();
			String msg = jibsMessages.convert("m_resume_match", obj);
			jibsServer.log(msg);

			board.getPlayerX().informPlayers(msg, playerO);
		} else {
			// m_resume_match_unlimited=%0 and %1 are resuming their unlimited
			// match.
			Object[] obj = { playerX.getName(), playerO.getName() };
			JibsMessages jibsMessages = jibsServer.getJibsMessages();
			String msg = jibsMessages.convert("m_resume_match_unlimited", obj);
			jibsServer.log(msg);

			board.getPlayerX().informPlayers(msg, playerO);
		}


		MoveChances nrChances = null;

		switch (board.getTurn()) {
		case 1:
			nrChances = new MoveChances(jibsServer, this, board);
			calcPossibleMovesAndOutBoard(this, board.getPlayerX(), true, board.getTurn(),
					nrChances, playerX, playerO, board, dice1.getValue(),
					dice2.getValue());

			break;

		case -1:
			nrChances = new MoveChances(jibsServer, this, board);
			calcPossibleMovesAndOutBoard(this, board.getPlayerX(), true, board.getTurn(),
					nrChances, playerX, playerO, board, dice1.getValue(),
					dice2.getValue());
		}
	}

	public int rollO(Player playerO, Player playerX,
			BackgammonBoard bgBoard, int turn) {
		Die dice11 = board.getoDie1();
		Die dice12 = board.getoDie2();
		dice11.roll();
		dice12.roll();
		board.setoDie1(dice11);
		board.setoDie2(dice12);

		// m_you_roll=You roll %0 and %1.
		Object[] obj = new Object[] {
				Integer.valueOf(dice11.getValue()),
				Integer.valueOf(dice12.getValue())
		};
		String msg = jibsServer.getJibsMessages().convert("m_you_roll", obj);
		playerO.getOutputStream().println(msg);
		// m_other_roll=%0 rolls %1 and %2.
		obj = new Object[] {
				playerO.getName(), Integer.valueOf(dice11.getValue()),
				Integer.valueOf(dice12.getValue())
		};
		msg = jibsServer.getJibsMessages().convert("m_other_roll", obj);
		showWatcherRoll(board.getPlayerX(), msg);

		Player tPlayer = board.getOpponent(playerO);
		tPlayer.getOutputStream().println(msg);

		MoveChances nrChances = new MoveChances(jibsServer, this, bgBoard);
		int nrMoves = calcPossibleMovesAndOutBoard(this, playerO, false, turn, nrChances,
				playerX, playerO, bgBoard, dice11.getValue(),
				dice12.getValue());

		return nrMoves;
	}

	public int rollX(Player playerX, Player playerO,
			BackgammonBoard bgBoard, int turn) {
		Die dice11 = board.getpDie1();
		Die dice12 = board.getpDie2();
		dice11.roll();
		dice12.roll();

		board.setpDie1(dice11);
		board.setpDie2(dice12);

		// m_you_roll=You roll %0 and %1.
		Object[] obj = new Object[] {
				Integer.valueOf(dice11.getValue()),
				Integer.valueOf(dice12.getValue())
		};
		String msg = jibsServer.getJibsMessages().convert("m_you_roll", obj);
		playerX.getOutputStream().println(msg);

		// DIK: Useless case statement?
		switch (board.getTurn()) {
		case 1:
			// m_other_roll=%0 rolls %1 and %2.
			obj = new Object[] {
					playerX.getName(), Integer.valueOf(dice11.getValue()),
					Integer.valueOf(dice12.getValue())
			};

			break;

		case -1:
			// m_other_roll=%0 rolls %1 and %2.
			obj = new Object[] {
					playerX.getName(), Integer.valueOf(dice11.getValue()),
					Integer.valueOf(dice12.getValue())
			};

			break;
		}

		msg = jibsServer.getJibsMessages().convert("m_other_roll", obj);

		Player tPlayer = board.getOpponent(playerX);
		tPlayer.getOutputStream().println(msg);
		showWatcherRoll(board.getPlayerX(), msg);

		MoveChances nrChances = new MoveChances(jibsServer, this, bgBoard);
		int nrMoves = calcPossibleMovesAndOutBoard(this, playerX, false, turn, nrChances,
				playerX, playerO, bgBoard, dice11.getValue(),
				dice12.getValue());

		return nrMoves;
	}

	public void save() {
		SqlMapClient sqlMap = jibsServer.getSqlMapInstance();

		try {
			if (true) {
				// delete old games
				JibsGame.deleteGames(jibsServer, board.getPlayerX(),
						board.getPlayerO());

				// save current game
				SavedGameParam savedGame = new SavedGameParam(this,
						board.getPlayerX(),
						board.getPlayerO(),
						getBackgammonBoard(),
						board.getMatchlength(),
						board.getTurn(),
						board.getMatchVersion(),
						new Date());
				sqlMap.insert("SavedGames.saveGame", savedGame);
				board.getPlayerX().endGame(board.getPlayerO());
				board.getPlayerO().endGame(board.getPlayerX());
				board.setEnded(true);
				board.getPlayerO().setGame(null);
				board.getPlayerX().setGame(null);
			}
		} catch (SQLException e) {
			jibsServer.logException(e);
		}
	}

	public void setBackgammonBoard(BackgammonBoard b) {
		this.board = b;
	}

	public void setJibsServer(JibsServer jibsServer) {
		this.jibsServer = jibsServer;
	}

	private void showWatcherRoll(Player player, String msg) {
		player.show2Watchers(msg);
	}

	public void startGame(int startTurn, Die dice1, Die dice2,
			int playerXPoints, int playerOPoints, int mayDouble1,
			int mayDouble2) {
		Player opponent = board.getOpponent(board.getPlayerX());

		board.setMayDoubleX(mayDouble1);
		board.setMayDoubleO(mayDouble2);
		board.setPlayerXScore(playerXPoints);
		board.setPlayerOScore(playerOPoints);
		board.setTurn(startTurn);

		MoveChances nrChances = null;

		nrChances = new MoveChances(jibsServer, this, board);

		int nrMoves = 0;

		switch (board.getTurn()) {
		case 1:
			nrMoves = calcPossibleMovesAndOutBoard(this, board.getPlayerO(), true,
					board.getTurn(), nrChances, playerX, playerO,
					board, dice1.getValue(),
					dice2.getValue());

			if (nrMoves > 0) {
				String nextCmd =  checkForcedMove(playerO, board,	nrMoves);
				if (nextCmd == null)
					nextCmd = checkGreedy(playerO, nrMoves);
				if (nextCmd != null)
					playerO.getClientWorker().executeCmd(nextCmd);
			} else {
				board.setTurn(-board.getTurn());

				Die die = new Die(jibsServer);
				die.setValue(0);
				board.setpDie1(die);
				board.setpDie2(die);
				opponent.getClientWorker().executeCmd("roll");
			}

			break;

		case -1:
			nrMoves = calcPossibleMovesAndOutBoard(this, board.getPlayerX(), true,
					board.getTurn(), nrChances, playerX, playerO,
					board, dice1.getValue(),
					dice2.getValue());

			if (nrMoves > 0) {
				String nextCmd =  checkForcedMove(playerO, board,	nrMoves);
				if (nextCmd == null)
					nextCmd = checkGreedy(playerO, nrMoves);
				if (nextCmd != null)
					playerX.getClientWorker().executeCmd(nextCmd);
			} else {
				board.setTurn(-board.getTurn());

				Die die = new Die(jibsServer);
				die.setValue(0);
				board.setpDie1(die);
				board.setpDie2(die);
				;
				opponent.getClientWorker().executeCmd("roll");
			}

			break;
		}
	}
// note: Axel calls X = player1 and O = player 2
	private final static int X = BackgammonBoard.X;
	private final static int O = BackgammonBoard.O;
	
	/** Player O has won the game.
	 * @param board
	 * @param winPoints
	 */
	public void winGameO(int winPoints) {
		Player winPlayer = board.getPlayerO();
		Player losePlayer = board.getPlayerX();

		winGameXO(O, winPoints, winPlayer, losePlayer);
	}

	public void winGameX(int winPoints) {
		Player winPlayer = board.getPlayerX();
		Player losePlayer = board.getPlayerO();
		winGameXO(X, winPoints, winPlayer, losePlayer);
	}

	private	int	not(int XO) {
		if (XO == X)
			return(O);
		return(X);
	}

	private void winGameXO(int XO, int winPoints,
			Player winPlayer, Player losePlayer) {
		JibsWriter winOut = winPlayer.getOutputStream();
		JibsWriter loseOut = losePlayer.getOutputStream();
		String msg = null;
		Object[] obj = null;
		// determine winPoints
		if (winPoints < 0) {
			int calc_winPoints = 1; // Normal

			if (board.getOnHome(not(XO)) <= 0) {
				calc_winPoints = 2; // Gammon
				if (XO == X) {
					for (int i = 19; i <= 24; i++) {
						if (board.getBoard()[i] > 0) {
							calc_winPoints = 3; // Backgammon
						}
					}
				} else {
					for (int i = 1; i <= 6; i++) {
						if (board.getBoard()[i] < 0) {
							calc_winPoints = 3; // Backgammon
						}
					}
				}
			}

			winPoints = calc_winPoints * board.getCubeNumber();
		}
		board.addPlayerScore(XO, winPoints);
		JibsMessages jibsMessages = jibsServer.getJibsMessages();

		obj = new Object[] {winPlayer.getName(), 
							Integer.valueOf(winPoints), 
							winPoints == 1 ? "" : "s"};
		msg = jibsMessages.convert("m_you_win_game", obj);
		winOut.println(msg);
		msg = jibsMessages.convert("m_other_win_game", obj);
		loseOut.println(msg);
		msg = jibsMessages.convert("m_watcher_win_game", obj);
		winPlayer.show2Watchers(msg);
		losePlayer.show2Watchers(msg);

		// XXX: Why are we outting a board here?
		BackgammonBoard opBoard = board.switch2O();
		if (board.isPlayerX(winPlayer)) {
			board.outBoard(winPlayer, "You", 0, 0, 0, 0, 0);
			opBoard.outBoard(losePlayer, "You", 0, 0, 0, 0, 0);
		} else {
			board.outBoard(losePlayer, "You", 0, 0, 0, 0, 0);
			opBoard.outBoard(winPlayer, "You", 0, 0, 0, 0, 0);
		}

		int winnerScore = board.getPlayerScore(XO);
		if ((winnerScore < board.getMatchlength())) {
		}

		int winScore = board.getPlayerXScore();
		String winName = board.getPlayerXName();
		int loseScore = board.getPlayerOScore();
		String loseName = board.getPlayerOName();

		if (loseScore > winScore) {
			winScore = board.getPlayerOScore();
			winName = board.getPlayerOName();
			loseScore = board.getPlayerXScore();
			loseName = board.getPlayerXName();
		}

		if (winnerScore < board.getMatchlength()) {
			obj = new Object[] { Integer.valueOf(board.getMatchlength()),
				winName, Integer.valueOf(winScore),
				loseName, Integer.valueOf(loseScore)
			};
			// m_score=score in %0 point match: %1-%2 %3-%4.
			msg = jibsMessages.convert("m_score", obj);
			winOut.println(msg);
			loseOut.println(msg);
			winPlayer.show2Watchers(msg);
			losePlayer.show2Watchers(msg);
			
			// match hasn't ended yet, start another game
			// m_join_next_game=Type 'join' if you want to play the next game,
			// type 'leave" if you don't.
			msg = jibsMessages.convert("m_join_next_game");
			winOut.println(msg);
			loseOut.println(msg);
		} else {
			Object[] obj6 = null;

			// m_match_over_all=%0 wins a %1 point match against %2 %3-%4 .
			obj6 = new Object[] {
					winPlayer.getName(),
					Integer.valueOf(board.getMatchlength()),
					losePlayer.getName(),
					Integer.valueOf(winScore),
					Integer.valueOf(loseScore)
			};

			msg = jibsMessages.convert("m_match_over_all", obj6);
			jibsServer.log(msg);
			winPlayer.informWatchers("m_match_over_all", obj6, true);
			losePlayer.informWatchers("m_match_over_all", obj6, true);

			// m_you_win_match=You win the %0 point match %1-%2 .
			obj6 = new Object[] {
					Integer.valueOf(board.getMatchlength()),
					Integer.valueOf(winScore),
					Integer.valueOf(loseScore)
			};

			String winMsg = jibsMessages.convert("m_you_win_match", obj6);

			// m_other_win_match=%0 wins the %1 point match %2-%3 .
			obj6 = new Object[] {
					winPlayer.getName(),
					Integer.valueOf(board.getMatchlength()),
					Integer.valueOf(winScore),
					Integer.valueOf(loseScore)
			};

			String loseMsg = jibsMessages.convert("m_other_win_match", obj6);

			int expWin = winPlayer.getExperience();
			double ratingWin = winPlayer.getRating();
			int expLose = losePlayer.getExperience();
			double ratingLose = losePlayer.getRating();
			double d = RatingCalculator.ratingChange(true,
					board.getMatchlength(),
					expWin, ratingWin, expLose,
					ratingLose);
			double newRating;
			int newExperience;

			newRating = ratingWin + d;
			newExperience = expWin + board.getMatchlength();

			winPlayer.changeRating(newRating, newExperience);

			newRating = ratingLose - d;
			newExperience = expLose + board.getMatchlength();

			losePlayer.changeRating(newRating, newExperience);
			winOut.println(winMsg);
			loseOut.println(loseMsg);
			new GameEnder(jibsServer, board.getPlayerX(), board.getPlayerO()).start();
		}
	}

	public void setCollection(ArrayList<PossibleMove> name) {
		this.col = name;
	}

	public Collection<PossibleMove> getCollection() {
		return this.col;
	}

	public Player getPlayerX() {
		return playerX;
	}

	public Player getPlayerO() {
		return playerO;
	}
}
